import 'package:after_layout/after_layout.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:osam_flutter/src/navigation/osam_route.dart';
import '../../osam_flutter.dart';

class OsamNavigator<P extends NavigationPresenter, T> extends StatefulWidget {
  final Map<T, dynamic> routes;
  final bool isMaterial;
  final bool log;
  final bool automaticHideKeyBoardOnPush;

  const OsamNavigator(
      {Key key, this.routes, this.isMaterial = false, this.log = false, this.automaticHideKeyBoardOnPush = false})
      : super(key: key);

  @override
  _OsamNavigatorState createState() => _OsamNavigatorState<P, T>();
}

class _OsamNavigatorState<P extends NavigationPresenter, T> extends State<OsamNavigator<P, T>> with AfterLayoutMixin {
  final _routes = <OsamRoute<dynamic, Object, Object>>[];
  var lastKnownStackSize = 0;

  Map<T, dynamic> get routes => widget.routes;

  P get presenter => context.presenter<P>();

  @override
  Widget build(BuildContext context) => routes[presenter.firstRoute.value](context);

  void silentPop() {
    if (_routes.length > 1) {
      lastKnownStackSize = lastKnownStackSize - 1;
      if (_routes.length > lastKnownStackSize) {
        _routes.removeLast();
        presenter.pop();
      }
    }
  }

  bool isWidgetBuilder(T route) => routes[route] is WidgetBuilder;

  void _hideKeyBoard() => SystemChannels.textInput.invokeMethod('TextInput.hide');

  PageRoute _pageRoute({WidgetBuilder widgetBuilder, RouteSettings settings}) => widget.isMaterial
      ? MaterialPageRoute(builder: widgetBuilder, settings: settings)
      : CupertinoPageRoute(builder: widgetBuilder, settings: settings);

  @override
  void afterFirstLayout(BuildContext context) {
    final navigator = Navigator.of(context);
    final initialRoutes = presenter.currentRoutes;
    _routes.addAll(initialRoutes);
    if (initialRoutes.length > 1) {
      final others = initialRoutes.sublist(1, initialRoutes.length);
      others.forEach((route) {
        final target = routes[route.value];
        if (!isWidgetBuilder(route.value)) {
          (target() as Future).then((_) {
            route.complete();
            silentPop();
          });
        } else {
          navigator.push(_pageRoute(widgetBuilder: target)).then((value) {
            route.complete();
            silentPop();
          });
        }
      });
      if (widget.automaticHideKeyBoardOnPush) _hideKeyBoard();
    }
    lastKnownStackSize = _routes.length;
    presenter.navigation.routes.addListener(() {
      final newRoutes = presenter.currentRoutes;
      if (widget.log) {
        print('Last routes: $_routes');
        print('New routes: $newRoutes');
      }
      final routesDifferentNumber = newRoutes.length - _routes.length;
      if (routesDifferentNumber == 0 && newRoutes.last != _routes.last) {
        final newRoute = newRoutes.last;
        final target = routes[newRoute.value];
        if (!isWidgetBuilder(newRoute.value)) {
          (target() as Future).then((_) {
            newRoute.complete();
            silentPop();
          });
        } else {
          navigator.pushReplacement(_pageRoute(widgetBuilder: target)).then((value) {
            newRoute.complete();
            silentPop();
          });
        }
        lastKnownStackSize = lastKnownStackSize + 1;
      } else if (routesDifferentNumber > 0) {
        final pushRoutes = newRoutes.sublist(_routes.length, newRoutes.length);
        pushRoutes.forEach((route) {
          if (!isWidgetBuilder(route.value)) {
            (routes[route.value]() as Future).then((_) {
              route.complete();
              silentPop();
            });
          } else {
            navigator.push(_pageRoute(widgetBuilder: routes[route.value])).then((_) {
              route.complete();
              silentPop();
            });
          }
        });
        _hideKeyBoard();
        lastKnownStackSize = newRoutes.length;
      } else if (routesDifferentNumber < 0) {
        if (_routes.first != newRoutes.first) {
          lastKnownStackSize = newRoutes.length;
          final newRoute = newRoutes.first;
          if (!isWidgetBuilder(newRoute.value)) throw Error();
          navigator
              .pushAndRemoveUntil(_pageRoute(widgetBuilder: routes[newRoute.value]), (route) => false)
              .then((value) {
            newRoute.complete();
            silentPop();
          });
        } else {
          _routes.sublist(newRoutes.length, _routes.length).forEach((route) {
            route.complete();
            navigator.pop();
          });
        }
      }
      _routes.clear();
      _routes.addAll(newRoutes);
    });
  }
}

class OsamTabNavigator<P extends TabNavigationPresenter, T> extends StatelessWidget {
  final Map<T, PresenterProvider<NavigationPresenter>> pages;
  final Map<T, Widget> tabs;
  final BoxDecoration tabDecoration;
  final Map<T, GlobalKey<NavigatorState>> navigatorKeys;

  const OsamTabNavigator({Key key, this.pages, this.tabDecoration, this.tabs, this.navigatorKeys}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    final presenter = context.presenter<P>();
    return ValueListenableBuilder(
        valueListenable: presenter.tabNavigation.tabNotifier,
        builder: (ctx, value, _) {
          final currentPresenter = pages[value].presenter;
          return Stack(
            children: <Widget>[
              IndexedStack(
                index: pages.keys.toList().indexOf(value),
                children: pages.keys.map((tab) => pages[tab]).toList(),
              ),
              Positioned(
                bottom: 0,
                child: SafeArea(
                  bottom: true,
                  child: Container(
                    decoration: tabDecoration,
                    padding: const EdgeInsets.symmetric(vertical: 5),
                    constraints: BoxConstraints(maxWidth: MediaQuery.of(context).size.width, minHeight: 48),
                    child: Row(
                      children: pages.keys
                          .map((tab) => Expanded(
                                child: GestureDetector(
                                  child: Container(
                                    child: tabs[tab],
                                    alignment: Alignment.center,
                                    color: Colors.transparent,
                                  ),
                                  onTap: () {
                                    if (value != tab) {
                                      presenter.changeTab(tab);
                                    } else {
                                      final routes = currentPresenter.currentRoutes;
                                      if (routes.length > 1) {
                                        currentPresenter.pushAndRemoveUntil(routes.first);
                                      }
                                    }
                                  },
                                ),
                              ))
                          .toList(),
                    ),
                  ),
                ),
              )
            ],
          );
        });
  }
}
